/* $Workfile: T2mSamples.c $       $Revision: 22 $      $Modtime: 9/01/05 2:26p $  */
/*******************************************************************************

  Odin TeleSystems Inc.

  OTX Telecom Adapter Driver

  @doc PublicDemo

  @module

    T2MSAMPLES.C

    A simple example console application demonstrating the use of OTX Driver
    with the following board combinations:
    
   The applications works with the following board combinations:

      Thor-2-PCI-Plus / Thor-4-PCI-Plus / Thor-8-PCI-Plus + Vidar-5x4-ASM-PRO,

    or Thor-2-PCMCIA-Pro,

    The Demo application provides a simple console interface which utilizes
    simple keystrokes as commands.

    It configures 2 E1 Lines and expects them to be looped with an external
    loop cable. 

    It configures 64 ConstVal Generator devices (one per timeslot), and 64 Val
    Sampler devices. The ConstVal generator device generates a constant value
    on a particular timeslot. The ValSampler devices are used to sample the value
    on a particular timeslot.

    Press 'h' or '?' for help. 'q' to quit.

*******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#if !defined(_WIN32_WCE)
#include <assert.h>
#include <time.h>
#endif  // #if !defined(_WIN32_WCE)

#include "OsaStd.h"
#include "OsaCon.h"
#include "OsaThrd.h"
#include "OtxGeneric.h"
#include "OtxErr.h"
#include "OtxDef.h"
#include "OtxType.h"
#include "OtxDrv.h"
#include "OtxHdlc.h"
#include "OtxRaw.h"
#include "OtxT1E1.h"
#include "OtxData.h"
#include "OtxTone.h"
#include "OtxDsp.h"
#include "OtxBrd.h"
#include "DspIoctl.h"

#define DEMO_DEFAULT_LOG_FILE "T2mSamples.log"
#define DEMO_DEFAULT_BOARD_NO 0

#define DEMO_TASK_REF 100L               // Task Reference for a non-blocking call

#define DEMO_DIALER_MAX 64

typedef struct DemoEnvT {
  char               *pszInputStr;
  OTX_BOOL            bExitOnError;
  char                cwd[OTX_MAX_PATH];
  char                logFileName[OTX_MAX_PATH];
  enum OtxPhysicalDeviceTypeE nBoardType;
  int                 coDsp;
  int                 boardNo;
  OtxT1E1LiModeE      eT1E1Mode;    // T1 or E1 mode
  OtxBrdClockSourceE  eClkSource;
  OtxT1LineCodeE      eT1LineCode;
  OtxT1FrameFormatE   eT1FrameFormat;
  OtxE1LineCodeE      eE1LineCode;
  OtxE1FrameFormatE   eE1FrameFormat;
  OtxDspSpmE          eProgramModule;      // Standard Program Module to be loaded in the DSP (from the Lib)
  OTX_BOOL            bUseDspDebugger;
} DemoEnvT;




/*******************************************************************************
 Function Prototypes
*******************************************************************************/
static void DemoCreateEventQueue(void);
static void DemoCreatePhysicalDevices(void);
static void DemoConfigureE1T1Devices(void);
static void DemoEnablePhysicalDevices(void);
static void DemoCreateLogicalDevices(void);
static void DemoConnectDevices(void);
static void DemoInitializeLogicalDevices(void);
static void DemoEnableLogicalDevices(void);
static void DemoInitT1E1ConfigData(OUT OtxE1ConfigOptionsS *pE1CfgData, OUT OtxT1ConfigOptionsS *pT1CfgData);
static void DemoPrintEvent(IN OtxEventDataS *pEventData);
static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP);
static void DemoHelpPanel(void);
static void DemoLogf(char *formatStr, ...);
static void DemoCheckRc(IN OTX_HANDLE hDevice, IN OTX_RESULT otxRc);
static void DemoExitQuestion(void);

static OTX_INT32 DemoProcessUserCommand(char commandChar);

/*******************************************************************************
 Global Data
*******************************************************************************/
OTX_EVENT   g_hNotificationEvent = OTX_INVALID_EVENT_VALUE;
OTX_HANDLE  g_hEventQueue  = 0;
FILE       *g_pDemoLogFile = NULL;
DemoEnvT    g_demoEnv;

static OsaMutexS g_analyzeCritSec;
static OsaMutexS g_demoLogfCritSec;


struct {
  OTX_HANDLE hThorController0;
  OTX_HANDLE hQuadE1T1[2];        // Only used for Thor-4-PCI-Plus and Thor-8-PCI(-Plus)
  OTX_HANDLE hVidarAsmController0;// not for Thor-2-PCMCIA series
  OTX_HANDLE hLi[2];
  OTX_HANDLE hDsp[4];
} g_hPhysDevs;


struct {
  OTX_HANDLE hPipe;
  OTX_HANDLE hHandle;
} g_sqConstValGenerator[64];

struct {
  OTX_HANDLE hPipe;
  OTX_HANDLE hHandle;
} g_sqValSampler[64];


struct {
  OtxE1ConfigOptionsS e1CfgData;
  OtxT1ConfigOptionsS t1CfgData;
} g_T1E1CfgData;



/*******************************************************************************
*******************************************************************************/


void DemoCreateEventQueue(void)
/*******************************************************************************

  @func

  Create the event queue to be used by the driver to inform the application
  about events, such as line status changes, or detected DMTF digits.

*******************************************************************************/
{
  OTX_RESULT result;

  // Create a notification Event
  result = OtxDrvCreateEvent(&g_hNotificationEvent); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);

  // Create Event Queue for driver notifications
  result = OtxDrvCreateEventQueue(128, g_hNotificationEvent, &g_hEventQueue); DemoCheckRc(g_hEventQueue, result);

  // Enable Event Queue
  result = OtxDrvEnable(g_hEventQueue, OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hEventQueue, result);

}



void DemoCreatePhysicalDevices(void)
/*******************************************************************************

  @func

  Create and Connect the Physical Devices on the Thor board

  @comm

    Calls function <f OtxDrvCreatePhysicalDevice> returning a handle to a
    Device Object representing a physical device, which can be the
    entire board or an integrated circuit (chip) on the board. The number of
    physical devices is set per board and cannot change dynamically.

    The physical devices can be classified in 2 categories:

  - BOARD Devices: A board device represents one OTX Network Interface card or
    a Daughter Board Module. The board device is the place holder and controller
    of other physical and sometime logical devices representing
    the devices on the board: such as Line Interface and Filter  Devices.


  - CHIP Devices: A chip device represents a physical integrated circuit
    on the board performing one or several functions. A logical device can
    be created to represent each of the functions implemented by a chip device.
    For example, a Digital Signal Processor is a physical "chip" device.
    However, the DSP software may implement several functions and be
    represented by a multiple of logical devices.



*******************************************************************************/
{
  OTX_RESULT  result;
  OTX_INT32   i;
  OTX_INT32   nQuadE1T1; // Only used for Thor-4-PCI-Plus, Thor-8-PCI(-Plus)

  // Connect to the Physical Board Controller
  switch (g_demoEnv.nBoardType) {
  case OTX_DEVICE_THOR_2_PCI_PLUS:
    DemoLogf("Opening Thor-2-PCI-PLUS board\n");
    break;
  case OTX_DEVICE_THOR_4_PCI_PLUS:
    DemoLogf("Opening Thor-4-PCI-Plus board\n");
    break;
  case OTX_DEVICE_THOR_8_PCI_PLUS:
    DemoLogf("Opening Thor-8-PCI-Plus board\n");
    break;
  case OTX_DEVICE_THOR_PCMCIA:
    DemoLogf("Opening OTX_DEVICE_THOR_PCMCIA board\n");
    break;
  case OTX_DEVICE_HUGIN_2_PC104:
    DemoLogf("Opening OTX_DEVICE_HUGIN_2_PC104 board\n");
    break;
  default:
    DemoLogf("Board Type (%d) not supported by this demo program\n", g_demoEnv.nBoardType);
    exit(1);
    break;
  }
  result = OtxDrvOpenPhysicalDevice(0, g_demoEnv.nBoardType, g_demoEnv.boardNo, g_demoEnv.boardNo, g_hEventQueue, &g_hPhysDevs.hThorController0); DemoCheckRc(g_hPhysDevs.hThorController0, result);


  
  switch (g_demoEnv.nBoardType) {
  case OTX_DEVICE_THOR_2_PCI_PLUS:
  case OTX_DEVICE_THOR_PCMCIA:
  case OTX_DEVICE_HUGIN_2_PC104:
    // On the Thor-2-PCI and the Thor-2-PCMCIA board the E1T1 devices are hosted by the board itself. Get handles to T1/E1 line interface devices #0 and #1
    for (i = 0; i < 2; i++) {
      DemoLogf("Open OTX_DEVICE_LI_T1E1 #%d\n", i);
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, OTX_DEVICE_LI_T1E1, i, i, g_hEventQueue, &g_hPhysDevs.hLi[i]); DemoCheckRc(g_hPhysDevs.hThorController0, result);
    }
    break;
  case OTX_DEVICE_THOR_4_PCI_PLUS:
    {
      // On the Thor-4-PCI-Plus board the E1T1 devices are hosted by a quad E1T1 device. 
      for (nQuadE1T1 = 0; nQuadE1T1 < 1; nQuadE1T1++) {
        DemoLogf("Open OTX_DEVICE_QUAD_T1E1 #%d\n", nQuadE1T1);
        result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, OTX_DEVICE_QUAD_T1E1, nQuadE1T1, nQuadE1T1, g_hEventQueue, &g_hPhysDevs.hQuadE1T1[nQuadE1T1]); DemoCheckRc(g_hPhysDevs.hThorController0, result);
      }
      
      // Get handles to T1/E1 line interface devices 
      for (i = 0; i < 2; i++) {
        nQuadE1T1 = i/4;
        DemoLogf("Open OTX_DEVICE_LI_T1E1 #%d (hosted by OTX_DEVICE_QUAD_T1E1 #%d)\n", i, nQuadE1T1);
        result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hQuadE1T1[nQuadE1T1], OTX_DEVICE_LI_T1E1, i%4, i, g_hEventQueue, &g_hPhysDevs.hLi[i]); DemoCheckRc(g_hPhysDevs.hQuadE1T1[nQuadE1T1], result);
      }
    }
    break;
  case OTX_DEVICE_THOR_8_PCI_PLUS:
    {
      // On the Thor-8-PCI and Thor-8-PCI-Plus board the E1T1 devices are hosted by a quad E1T1 device. Open those two devices.
      for (nQuadE1T1 = 0; nQuadE1T1 < 2; nQuadE1T1++) {
        DemoLogf("Open OTX_DEVICE_QUAD_T1E1 #%d\n", nQuadE1T1);
        result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, OTX_DEVICE_QUAD_T1E1, nQuadE1T1, nQuadE1T1, g_hEventQueue, &g_hPhysDevs.hQuadE1T1[nQuadE1T1]); DemoCheckRc(g_hPhysDevs.hThorController0, result);
      }
      
      // Get handles to T1/E1 line interface devices 
      for (i = 0; i < 2; i++) {
        nQuadE1T1 = i/4;
        DemoLogf("Open OTX_DEVICE_LI_T1E1 #%d (hosted by OTX_DEVICE_QUAD_T1E1 #%d)\n", i, nQuadE1T1);
        result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hQuadE1T1[nQuadE1T1], OTX_DEVICE_LI_T1E1, i%4, i, g_hEventQueue, &g_hPhysDevs.hLi[i]); DemoCheckRc(g_hPhysDevs.hQuadE1T1[nQuadE1T1], result);
      }
    }
    break;
  default:
    DemoLogf("Board Type (%d) not supported by this demo program\n", g_demoEnv.nBoardType);
    exit(1);
    break;
  }
  


  // Open physical DSP devices
  switch (g_demoEnv.nBoardType) {
  case OTX_DEVICE_THOR_PCMCIA:
  case OTX_DEVICE_HUGIN_2_PC104:
    // Get handles to the physcial DSP devices #0 and #1
    for (i = 0; i < g_demoEnv.coDsp; i++) {
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, OTX_DEVICE_DSP, i, i, g_hEventQueue, &g_hPhysDevs.hDsp[i]); DemoCheckRc(g_hPhysDevs.hThorController0, result);
    }
    break;
  case OTX_DEVICE_THOR_2_PCI_PLUS:
  case OTX_DEVICE_THOR_4_PCI_PLUS:
  case OTX_DEVICE_THOR_8_PCI_PLUS:
    {
      OtxPhysicalDeviceTypeE eDevType;
      OTX_ITERATOR itrChildTypes;
      OTX_BOOL bOpenAsm = OTX_FALSE;
      OtxDrvIteratorBegin(&itrChildTypes);
      while (OTX_SUCCESS(OtxDrvEnumDeviceType(g_hPhysDevs.hThorController0, &itrChildTypes, (OTX_INT32 *)&eDevType))) {  
        switch (eDevType) {
        case OTX_DEVICE_VIDAR_5X4_ASM_PRO:
          DemoLogf("Opening Vidar-5x4-ASM-PRO board\n");
          bOpenAsm = OTX_TRUE;
          break;
        case OTX_DEVICE_VIDAR_55x4_ASM:
          DemoLogf("Opening Vidar-55x4-ASM board\n");
          bOpenAsm = OTX_TRUE;

          // Verify that the system supports the Vidar-55x4-ASM board
          {
            char szCpldBuildNumbers[256];
            char szBrdBuildNumber[16];
            OTX_INT32 nBoardBuildNo;
          
            if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_PCI) {
            
              //  Verify firmware revision
              result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_CPLD_BUILD, szCpldBuildNumbers, 256); DemoCheckRc(g_hPhysDevs.hThorController0, result);
              if (strstr(szCpldBuildNumbers, "OTXCDC: 0761") || strstr(szCpldBuildNumbers, "OTXCDC: 0762")) {
                DemoLogf("Error: The firmware on the Thor-2-PCI must be upgraded to firmware build 5\n");
                DemoLogf("       (OTXCDC 0764) or later to support the Vidar-55x4-ASM board.\n");
                DemoLogf("       Please download firmware update program from the Odin TeleSystems\n");
                DemoLogf("       web site (www.odints.com)\n");
                DemoLogf("       Also please verify that the HMA-1098-1-2.0 module is plugged into connector BJ3.\n");
                DemoExitQuestion();
              }
            
              // Verify board build number
              result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_BUILD, szBrdBuildNumber, 16); DemoCheckRc(g_hPhysDevs.hThorController0, result);
              nBoardBuildNo = strtol(strtok(szBrdBuildNumber, "."), NULL, 0);
              DemoLogf("Board Build Number: %d\n", nBoardBuildNo);
              if (nBoardBuildNo < 11) {
                DemoLogf("Error: The Thor-2-PCI board must be upgraded to build 11\n");
                DemoLogf("       or later to support the Vidar-55x4-ASM board.\n");
                DemoLogf("       Please contact Odin TeleSystems Support Team for upgrade details.\n");
                DemoExitQuestion();
              }
            
            } else if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_8_PCI) {
              //  Verify firmware revision
              result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_CPLD_BUILD, szCpldBuildNumbers, 256); DemoCheckRc(g_hPhysDevs.hThorController0, result);
              if (strstr(szCpldBuildNumbers, "OTXCDC: 0771")) {
                DemoLogf("Error: The firmware on the Thor-8-PCI must be upgraded to firmware build 3\n");
                DemoLogf("       (OTXCDC 0772) or later to support the Vidar-55x4-ASM board.\n");
                DemoLogf("       Please download firmware update program from the Odin TeleSystems\n");
                DemoLogf("       web site (www.odints.com)\n");
                DemoLogf("       Also please verify that the HMA-1098-1-2.0 module is plugged into connector BJ3.\n");
                DemoExitQuestion();
              }
            
              // Verify board build number
              result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_BUILD, szBrdBuildNumber, 16); DemoCheckRc(g_hPhysDevs.hThorController0, result);
              nBoardBuildNo = strtol(szBrdBuildNumber, NULL, 0);
              DemoLogf("Board Build Number: %d\n", nBoardBuildNo);
              if (nBoardBuildNo < 5) {
                DemoLogf("Error: The Thor-8-PCI board must be upgraded to build 5\n");
                DemoLogf("       or later to support the Vidar-55x4-ASM board.\n");
                DemoLogf("       Please contact Odin TeleSystems Support Team for upgrade details.\n");
                DemoExitQuestion();
              }
            
            }
          
          }
          break;

        default:
          break;

        }

        if (bOpenAsm) break;

      }

      if (bOpenAsm) {
        // Get a handle to to the ASM Daughter Board
        result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, eDevType, 0, 0, g_hEventQueue, &g_hPhysDevs.hVidarAsmController0); DemoCheckRc(g_hPhysDevs.hThorController0, result);
        for (i = 0; i < g_demoEnv.coDsp; i++) {
          result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hVidarAsmController0, OTX_DEVICE_DSP, i, i, g_hEventQueue, &g_hPhysDevs.hDsp[i]); DemoCheckRc(g_hPhysDevs.hVidarAsmController0, result);
        }
      } else {
        DemoLogf("The board is not populated with the required Vidar-ASM board.\n");
        DemoExitQuestion();
      }
      
    }
    break;
  default:
    DemoLogf("Board Type (%d) not supported by this demo program\n", g_demoEnv.nBoardType);
    exit(1);
    break;
  }
  
  
  // Load DSP Program
  for (i = 0; i < g_demoEnv.coDsp; i++) {
    DemoLogf("DSP#%d: Loading DSP program OTX_DSP_SPM_OTXSPM2\n", i);
    result = OtxDspRunSpm(g_hPhysDevs.hDsp[i], g_demoEnv.eProgramModule); DemoCheckRc(g_hPhysDevs.hDsp[i], result);
  }

  if (g_demoEnv.bUseDspDebugger) {
    DemoLogf("Please load and run DSP to DSP#0 and DSP#1 in your DSP Debugger\n");
    DemoLogf("Press any key when the program is running...\n");
    OsaConsoleGetCh();
  }

}



/*******************************************************************************

  @func
    Configure E1/T1 Devices

  @comm The clock source of the board(s) must set by a call 
  to <f OtxBrdSetClocks> prior to calling this function.

*******************************************************************************/
void DemoConfigureE1T1Devices(void)
{
  OTX_RESULT result;
  OTX_INT32  i;

  // Set default configuration data
  DemoInitT1E1ConfigData( &(g_T1E1CfgData.e1CfgData), &(g_T1E1CfgData.t1CfgData));
  
  for (i = 0; i < 2; i++) {

    // Set termination for the T1/E1 Front End
    result = OtxT1E1SetLineTermination(g_hPhysDevs.hLi[i], OTX_T1E1_TERM_100_OHM); DemoCheckRc(g_hPhysDevs.hLi[i], result);

    // Configure Transceivers
    if (g_demoEnv.eT1E1Mode == OTX_T1E1_MODE_T1) {
      // T1 mode
      result = OtxT1SetConfiguration(g_hPhysDevs.hLi[i], &(g_T1E1CfgData.t1CfgData));  DemoCheckRc(g_hPhysDevs.hLi[i], result);
    } else {
      // E1 mode
      result = OtxE1SetConfiguration(g_hPhysDevs.hLi[i], &(g_T1E1CfgData.e1CfgData));  DemoCheckRc(g_hPhysDevs.hLi[i], result);
    }
    result = OtxT1E1SetLiMode(g_hPhysDevs.hLi[i], g_demoEnv.eT1E1Mode); DemoCheckRc(g_hPhysDevs.hLi[i], result);

  }

}



void DemoEnablePhysicalDevices(void)
/*******************************************************************************

  @func
    Enable Configured Physical Devices

*******************************************************************************/
{
  OTX_RESULT  result;
  OTX_INT32   i;
  OTX_INT32   nQuadE1T1; // Only used for Thor-4-PCI-Plus, Thor-8-PCI-(Plus)

  // Enable the board (Physical Device)
  DemoLogf("Enable hThorController0\n");
  result = OtxDrvEnable(g_hPhysDevs.hThorController0,  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hThorController0, result);

  if (g_hPhysDevs.hVidarAsmController0) {
    result = OtxDrvEnable(g_hPhysDevs.hVidarAsmController0,  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hVidarAsmController0, result);
  }

    // If board type is Thor-4-PCI-Plus then enable Quad E1T1 device
  if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_4_PCI_PLUS) {
      
    // Enable Quad E1T1 devices
    for (nQuadE1T1 = 0; nQuadE1T1 < 1; nQuadE1T1++) {
      DemoLogf("Enable OTX_DEVICE_QUAD_T1E1 #%d device\n", nQuadE1T1);
      result = OtxDrvEnable(g_hPhysDevs.hQuadE1T1[nQuadE1T1], OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hQuadE1T1[nQuadE1T1], result);
    }
  }

  // If board type is Thor-8-PCI-Plus then enable Quad E1T1 device
  if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_8_PCI_PLUS) {

    // Enable Quad E1T1 devices
    for (nQuadE1T1 = 0; nQuadE1T1 < 2; nQuadE1T1++) {
      DemoLogf("Enable OTX_DEVICE_QUAD_T1E1 #%d device\n", nQuadE1T1);
      result = OtxDrvEnable(g_hPhysDevs.hQuadE1T1[nQuadE1T1], OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hQuadE1T1[nQuadE1T1], result);
    }
  }


  for (i = 0; i < 2; i++) {

    // Enable Physical Devices
    DemoLogf("Enable g_hPhysDevs.hLi[%d]\n", i);
    result = OtxDrvEnable(g_hPhysDevs.hLi[i],  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hLi[i], result);

    DemoLogf("g_hPhysDevs.hDsp[%d]\n", i);
    result = OtxDrvEnable(g_hPhysDevs.hDsp[i], OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hDsp[i], result);

  }

}


void DemoCreateLogicalDevices(
/*******************************************************************************

  @func
    Create the needed Logical Devices

  @comm
    The logical devices are differenciated from the phyiscal devices by the fact
    the number of logical devices is not set. More of them can be requested
    by the user and allocated by the system if needed.

    Logical Devices can be classified into Control Device and Data Devices:

    - Control Device:  A logical device performing a control function. E.g. Event Queue.

    - Data Filter Device: A device implemented in Hardware, Firmware of Software
    which converts, processes, or manipulates serial data. Examples of data devices include
    Time-Space Switch, HDLC Controller, Tone Detector, Tone Generator, etc.

    Data Filter devices can be categorized into 3 basic types:

     - Data Source: Originates a Pipe or adds data to an existing Pipe. E.g. DTMF Tone Generator.

     - Data Sink: Terminates a Pipe or extracts data from a Pipe. E.g. DTMF Detector.

     - Data Bridge: Cross-connects data from one Pipe to another or combines data from
     multiple Pipes into one Pipe. E.g. cross-connect in TSS.

*******************************************************************************/
  void
)
{
  OTX_RESULT result;
  OTX_INT32 i;
  OTX_INT32 j;

  // Create 64 ConstVal Generators and 64 ValSamples (32 on DSP#0 and 32 on DSP#1)
  for (i = 0; i < 2; i++) {
    for (j = 0; j < 32; j++) {
      
      //DemoLogf("Creating logical device [%d]n", i*32 + j);
      
      // Create Const Val Generator
      result = OtxDrvCreateLogicalDevice(g_hPhysDevs.hDsp[i], OTX_LDEVICE_TONE_CONST_VAL_GENERATOR, g_hEventQueue, i*32 + j, &g_sqConstValGenerator[i*32 + j].hHandle); DemoCheckRc(g_hPhysDevs.hDsp[i], result);
      // Create Val Sampler
      result = OtxDrvCreateLogicalDevice(g_hPhysDevs.hDsp[i], OTX_LDEVICE_TONE_VAL_SAMPLER, g_hEventQueue, i*32 + j, &g_sqValSampler[i*32 + j].hHandle); DemoCheckRc(g_hPhysDevs.hDsp[i], result);
      
    }
  }  
  
}


void DemoConnectDevices(
/*******************************************************************************

  @func
    Create the needed Pipe Devices and connect the physical and logical devices
    to pipes.

  @comm
    A Pipe Device is a logical entity representing a serial bit-stream (a subset of a
    Highway) with a data rate between 8 kbit/s and 2.048Mbit/s. A pipe is a characterized
    bu its capacity (in 8 kbit/s increments). The user can create a pipe with
    the driver function <f OtxDrvCreatePipe>. A pipe is connected to physical or
    logical devices and mapped to specific time-slots with the driver function
    <f OtxDrvConnectPipe>.

*******************************************************************************/
  void
)
{
  OTX_INT32 ts;
  OTX_INT32 nCapacity;
  OTX_INT32 i;
  OTX_RESULT result;
  OTX_UINT8 sqTimeSlotMask[32];
  OTX_INT32 j;
  
  
  // CONNECT LOGICAL DEVICES WITHIN THE DSP
  // --------------------------------------
  for (i = 0; i < 2; i++) {
    
    for (j = 0; j < 32; j++) {
      
      //DemoLogf("Connecting Logical device [%d]\n", i*32 + j);
      
      // Receive direction:
      {
        //DemoLogf("Connect DSP Val Samplers\n");
        ts = j;
        memset(sqTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
        sqTimeSlotMask[ts] = 0xFF;
        nCapacity = 64;
        
        result = OtxDrvCreatePipe(nCapacity, &g_sqValSampler[i*32 + j].hPipe); DemoCheckRc(g_sqValSampler[i*32 + j].hPipe, result);
        result = OtxDrvConnectPipe(g_sqValSampler[i*32 + j].hPipe, OTX_PIPE_OUTPUT,  g_hPhysDevs.hDsp[i], -1, sqTimeSlotMask, 32); DemoCheckRc(g_sqValSampler[i*32 + j].hPipe, result);
        result = OtxDrvConnectPipe(g_sqValSampler[i*32 + j].hPipe, OTX_PIPE_INPUT,   g_hPhysDevs.hLi[i],  -1, sqTimeSlotMask, 32); DemoCheckRc(g_sqValSampler[i*32 + j].hPipe, result);
        
        result = OtxDrvConnectLogicalDevice(g_sqValSampler[i*32 + j].hPipe, -1, g_sqValSampler[i*32 + j].hHandle, -1); DemoCheckRc(g_sqValSampler[i*32 + j].hPipe, result);
      }
      
      // Transmit direction
      {
        //DemoLogf("Connect ConstVal Generators\n");
        ts = j;
        memset(sqTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
        sqTimeSlotMask[ts] = 0xFF;
        nCapacity = 64;
        
        result = OtxDrvCreatePipe(nCapacity, &g_sqConstValGenerator[i*32 + j].hPipe); DemoCheckRc(g_sqConstValGenerator[i*32 + j].hPipe, result);
        
        result = OtxDrvConnectPipe(g_sqConstValGenerator[i*32 + j].hPipe, OTX_PIPE_INPUT,  g_hPhysDevs.hDsp[i], -1, sqTimeSlotMask, 32); DemoCheckRc(g_sqConstValGenerator[i*32 + j].hPipe, result);
        result = OtxDrvConnectPipe(g_sqConstValGenerator[i*32 + j].hPipe, OTX_PIPE_OUTPUT, g_hPhysDevs.hLi[i],  -1, sqTimeSlotMask, 32); DemoCheckRc(g_sqConstValGenerator[i*32 + j].hPipe, result);
        
        result = OtxDrvConnectLogicalDevice(g_sqConstValGenerator[i*32 + j].hHandle,  -1, g_sqConstValGenerator[i*32 + j].hPipe,  -1); DemoCheckRc(g_sqConstValGenerator[i*32 + j].hHandle, result);
      }
      
    } // j
  } // i
  
  
}


void DemoInitializeLogicalDevices(
/*******************************************************************************

  @func
    Initialize the logical device. This function must be called after 
    the devices are created and connected.

*******************************************************************************/
  void
)
{
  OTX_RESULT result;
  OTX_INT32 i;
  OTX_INT32 j;

  for (i = 0; i < 2; i++) {
    
    // Initialize the hosted in the DSP
    for (j = 0; j < 32; j++) {
      result = OtxToneConstValGeneratorInit(g_sqConstValGenerator[i*32 + j].hHandle, OTX_DATA_FORMAT_UINT8, 0x55); DemoCheckRc(g_sqConstValGenerator[i*32 + j].hHandle, result);
      result = OtxToneValSamplerInit(g_sqValSampler[i*32 + j].hHandle, OTX_DATA_FORMAT_UINT8); DemoCheckRc(g_sqValSampler[i*32 + j].hHandle, result);
    }
  }

}


void DemoEnableLogicalDevices(void)
/*******************************************************************************

  @func
    Enable Configured Logical Devices

*******************************************************************************/
{
  OTX_RESULT result;
  OTX_INT32 i;
  OTX_INT32 j;

  for (i = 0; i < 2; i++) {
    for (j = 0; j < 32; j++) {
      //DemoLogf("Enable hValSampler[%d]\n", i*32 + j);
      result = OtxDrvEnable(g_sqValSampler[i*32 + j].hHandle,  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_sqValSampler[i*32 + j].hHandle, result);
    }
  }

}


void DemoInitT1E1ConfigData(
/*******************************************************************************

  @func
    Initialize the T1/E1 Configuration Data Structures passed to the T1/E1
    Line Interface physical devices

*******************************************************************************/
  OUT OtxE1ConfigOptionsS *pE1CfgData,
  OUT OtxT1ConfigOptionsS *pT1CfgData
)
{
  // Initialize E1 Configuration Data
  pE1CfgData->eTxLineCode                = g_demoEnv.eE1LineCode;
  pE1CfgData->eRxLineCode                = g_demoEnv.eE1LineCode;
  pE1CfgData->eTxFrameFormat             = g_demoEnv.eE1FrameFormat;
  pE1CfgData->eRxFrameFormat             = g_demoEnv.eE1FrameFormat;
  pE1CfgData->eAisDetection              = OTX_E1_AIS_ETS300233;
  pE1CfgData->bfSiBits                   = 0x3;
  pE1CfgData->bfSaBits                   = 0x1F;
  pE1CfgData->bExtendedHDB3errorDetect   = OTX_FALSE;
  pE1CfgData->bAutoRegainMultiframe      = OTX_FALSE;
  pE1CfgData->bAutoResync                = OTX_TRUE;
  pE1CfgData->bAutoCrcBitInversion       = OTX_FALSE;
  pE1CfgData->bAutoRemoteAlarm           = OTX_FALSE;
  pE1CfgData->nLosSensitivity            = 10;  // 10 = LOS detection after 176 consecutive zeros which fulfills G.775
  pE1CfgData->nLosRecovery               = 21;  // 21 = LOS recovery after 22 ones in the LosRecovery interval which fulfills G.775
  pE1CfgData->nLineLength                = 10;
  pE1CfgData->bHighTxPower               = OTX_FALSE;
  pE1CfgData->bReceiveEqualizer          = OTX_FALSE;

  // Initialize T1 Configuration Data
  pT1CfgData->eSignalingMode             = OTX_T1_SM_CCS;
  pT1CfgData->eTxLineCode                = g_demoEnv.eT1LineCode;
  pT1CfgData->eRxLineCode                = g_demoEnv.eT1LineCode;
  pT1CfgData->eFrameFormat               = g_demoEnv.eT1FrameFormat;
  pT1CfgData->bCRC6Enable                = OTX_TRUE;
  pT1CfgData->eTxRemoteAlarmFormat       = OTX_T1_RA_YELLOW_A;
  pT1CfgData->eRxRemoteAlarmFormat       = OTX_T1_RA_YELLOW_A;
  pT1CfgData->bAutoResync                = OTX_TRUE;
  pT1CfgData->bAutoCrcBitInversion       = OTX_FALSE;
  pT1CfgData->eLfaSensitivity            = OTX_T1_LFA_2_OUT_OF_4;
  pT1CfgData->bAutoRemoteAlarm           = OTX_FALSE;
  pT1CfgData->nLosSensitivity            = 10;  // 10 = LOS detection after 176 consecutive zeros which fulfills G.775, Telcordia (Bellcore) and AT&T
  pT1CfgData->nLosRecovery               = 21;  // 21 = LOS recovery after 22 ones in the LosRecovery interval which fulfills G.775, Telcordia (Bellcore) and AT&T
  pT1CfgData->nLineLength                = 10;
  pT1CfgData->bHighTxPower               = OTX_FALSE;
  pT1CfgData->bReceiveEqualizer          = OTX_FALSE;
  pT1CfgData->bfClearChannels            = 0x000000; // 0x000000 = No channels are marked as clear channels
}





/*******************************************************************************
  @func
    Idle Function that check for Events from the OTX Driver
*******************************************************************************/
void *DemoIdleFunction(
  void *lpParameter   // thread data
)
{
  OtxEventDataS eventData;
  OTX_RESULT result;

  while (1) {
    OtxDrvWaitForSingleEvent(g_hNotificationEvent, OTX_TIME_INFINITY);
    do {
      result = OtxDrvGetEventData(g_hEventQueue, &eventData); DemoCheckRc(g_hEventQueue, result);

      if (OTX_RESULT_CODE(result) == OTX_S_OK) {
        // We have a driver event
        // Print the event data.
        // In a real application branch based on the data
        DemoPrintEvent(&eventData);

      }
    } while (OTX_RESULT_CODE(result) == OTX_S_OK);
  }

  return 0;

}


void DemoPrintEvent(
/*******************************************************************************
  @func
    Prints a Driver Event. If HDLC Data, reads the data.
*******************************************************************************/
  OtxEventDataS *pEventData
)
{
  OTX_RESULT otxRc;
  char msgStr[256];

  //OsaMutexLock(&g_analyzeCritSec);

  switch (pEventData->m_eDeviceType) {
  case OTX_DEVICE_THOR_2_PCI_PLUS:
  case OTX_DEVICE_THOR_PCMCIA:
  case OTX_DEVICE_HUGIN_2_PC104:
  case OTX_DEVICE_THOR_4_PCI_PLUS:
  case OTX_DEVICE_THOR_8_PCI_PLUS:
    // Event from the board
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      if (pEventData->m_nParam) {
        DemoLogf("Brd[%d] Event: %s (%d)\n", pEventData->m_nSourceId, msgStr, pEventData->m_nParam);
      } else {
        DemoLogf("Brd[%d] Event: %s\n", pEventData->m_nSourceId, msgStr);
      }
    } else {
      DemoLogf("Brd[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }

    break;

  case OTX_DEVICE_LI_T1E1:
    // Event from Li0 or Li1
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      if (pEventData->m_nParam) {
        DemoLogf("E1/T1[%d] Event: %s (%d)\n", pEventData->m_nSourceId, msgStr, pEventData->m_nParam);
      } else {
        DemoLogf("E1/T1[%d] Event: %s\n", pEventData->m_nSourceId, msgStr);
      }
    } else {
      DemoLogf("E1/T1[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    
    break;

  case OTX_DEVICE_DSP:
    // Non-logical device related event from the DSPs
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      DemoLogf("DSP[%d] Event: %s \n", pEventData->m_nSourceId, msgStr);
    } else {
      DemoLogf("DSP[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    break;
    
  default:
    // Unhandled event
    DemoLogf("Event from an unknown object 0x%X\n", pEventData->m_hDevice);
    break;
  }


  //OsaMutexUnlock(&g_analyzeCritSec);

}






/*******************************************************************************

  @func

  Parse the program's command line arguments and calls upon
  requested actions.

  There are some optional switches that can be used following
  the driver name on the command line. Switches are are
  separated by spaces, and can appear in arbitrary order.
  Switches consist of a hyphen and a letter:

  @comm

*******************************************************************************/
static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP) {
  short  i;

  // Set reasonable defaults before reading from command line
  envP->pszInputStr       = NULL;
  envP->bExitOnError      = OTX_FALSE;
  envP->boardNo           = DEMO_DEFAULT_BOARD_NO;
  envP->eT1E1Mode         = OTX_T1E1_MODE_E1;
  envP->eClkSource        = OTX_CLOCK_SOURCE_LOCAL_0;
  envP->eProgramModule    = OTX_DSP_SPM_OTXSPM2;
  envP->nBoardType        = OTX_DEVICE_THOR_PCMCIA;
  envP->bUseDspDebugger   = OTX_FALSE;
  envP->coDsp             = 2;
  envP->eT1LineCode       = OTX_T1_LC_B8ZS;
  envP->eT1FrameFormat    = OTX_T1_FF_ESF;
  envP->eE1LineCode       = OTX_E1_LC_HDB3;
  envP->eE1FrameFormat    = OTX_E1_FF_CRC4_MULTIFRAME;


  strcpy(envP->logFileName, envP->cwd);
  strcat(envP->logFileName, DEMO_DEFAULT_LOG_FILE);
  for (i = 1; i < argc; i++) {        // Scan command line arguments
    if (*(argv[i]+1) == 'Z') {   // Command input string -- case sensitive
      envP->pszInputStr = argv[i]+2;
    } else {    
      OsaStdStrUpr(argv[i]);
      if (*argv[i] == '-' || *argv[i] == '/') {  // Is argument a legal switch?
        if (*(argv[i]+1) == 'L') {   // Log file
          strcpy(envP->logFileName, argv[i] + 2);
        } else if (*(argv[i]+1) == 'Q') {   // Exit on error
          envP->bExitOnError = OTX_TRUE;
        } else if (*(argv[i]+1) == 'B') {   // Board number
          envP->boardNo = strtoul(argv[i]+2, NULL, 0);
        } else if (*(argv[i]+1) == 'D') {   // DSP Debugger
          envP->bUseDspDebugger = OTX_TRUE;
          envP->coDsp             = 4;
        } else if (*(argv[i]+1) == 'T') {   // Board Type
          if ( *(argv[i]+2) == '4') {
            envP->nBoardType  = OTX_DEVICE_THOR_4_PCI_PLUS;
          } else if ( *(argv[i]+2) == 'T') {
            envP->nBoardType  = OTX_DEVICE_THOR_2_PCI_PLUS;
          } else if ( *(argv[i]+2) == 'M') {
            envP->nBoardType  = OTX_DEVICE_THOR_PCMCIA;
          //} else if ( *(argv[i]+2) == 'H') {
          //  envP->nBoardType  = OTX_DEVICE_HUGIN_2_PC104;
          } else if ( *(argv[i]+2) == 'P') {
            envP->nBoardType  = OTX_DEVICE_THOR_8_PCI_PLUS;
          } else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            exit(1);
          }
        
        } else if (*(argv[i]+1) == 'K') {   // Clock Source
          if (*(argv[i]+2) == 'I') {   // Internal
            envP->eClkSource = OTX_CLOCK_SOURCE_INTERNAL;
          } else if (*(argv[i]+2) == '0') {   // Line Interface 0
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_0;
          } else if (*(argv[i]+2) == '1') {   // Line Interface 1
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_1;
          } else if (*(argv[i]+2) == 'E') {   // External
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_2;
          } else if (*(argv[i]+2) == '3') {   // Split clock, both Lis synchronized to its own line
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_3;
          } else if (*(argv[i]+2) == '4') {   // Split clock, Li0 synchronized to its T1E1 span. Li1 internal
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_4;
          } else if (*(argv[i]+2) == '5') {   // Split clock, Li1 synchronized to its T1E1 span. Li0 internal
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_5;
          } else {
            DemoLogf("Error: Invalid Clock Source command line switch %s\n", argv[i]);
            DemoHelpPanel();
            exit(1);
          }
        } else if (*(argv[i]+1) == 'M') {   // Li Mode
          if (*(argv[i]+2) == 'E') {   // E1 
            envP->eT1E1Mode    = OTX_T1E1_MODE_E1;
          } else if (*(argv[i]+2) == 'T') {   // T1
            envP->eT1E1Mode    = OTX_T1E1_MODE_T1;
          } else {
            DemoLogf("Error: Invalid Li Mode (E1/T1) command line switch %s\n", argv[i]);
            DemoHelpPanel();
            exit(1);
          }
        } else if (*(argv[i]+1) == 'F') {   // T1 Frame format
          if ( *(argv[i]+2) == 'E') {
            envP->eT1FrameFormat = OTX_T1_FF_ESF; // ESF
          } else if ( *(argv[i]+2) == '4') {
            envP->eT1FrameFormat = OTX_T1_FF_F4; 
          } else if ( *(argv[i]+2) == 'S') {
            envP->eT1FrameFormat = OTX_T1_FF_F12; // SF
          } else if ( *(argv[i]+2) == '9') {
            envP->eT1FrameFormat = OTX_T1_FF_F72; // SLC96
          } else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
          
        } else if (*(argv[i]+1) == 'J') {   // T1 Line Code
          if ( *(argv[i]+2) == 'A') {
            envP->eT1LineCode = OTX_T1_LC_AMI; // AMI
          } else if ( *(argv[i]+2) == 'B') {
            envP->eT1LineCode = OTX_T1_LC_B8ZS; // B8ZS
          } else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
          
        } else if (*(argv[i]+1) == 'G') {   // E1 Frame format
          if ( *(argv[i]+2) == 'D') {
            envP->eE1FrameFormat = OTX_E1_FF_DOUBLE_FRAME; 
          } else if ( *(argv[i]+2) == 'C') {
            envP->eE1FrameFormat = OTX_E1_FF_CRC4_MULTIFRAME; 
          } else if ( *(argv[i]+2) == 'G') {
            envP->eE1FrameFormat = OTX_E1_FF_CRC4_MULTIFRAME_G706; 
          } else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
          
        } else if (*(argv[i]+1) == 'U') {   // E1 Line Code
          if ( *(argv[i]+2) == 'A') {
            envP->eE1LineCode = OTX_E1_LC_AMI; // AMI
          } else if ( *(argv[i]+2) == 'H') {
            envP->eE1LineCode = OTX_E1_LC_HDB3; // HDB3
          } else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
          
        } else {
          DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
          DemoHelpPanel();
          exit(1);
        }
      } else {
        DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
        DemoHelpPanel();
        exit(1);
      }
    }
  }

}


/*******************************************************************************

  @func

  Displays a help for the command line.

  @comm

*******************************************************************************/
static void DemoHelpPanel()
{
  DemoLogf("\nCommand line switches:\n");
  DemoLogf(" /B<boardNo>   : Board number (Default is %d)\n", DEMO_DEFAULT_BOARD_NO);
  DemoLogf(" /T[T|4|P|M]     : Board Type used for the demo, where\n");
  DemoLogf("                   2 = Use Thor-2-PCI-Plus Board with Vidar-5x4-ASM-PRO \n");
  DemoLogf("                   4 = Use Thor-4-PCI-Plus Board with Vidar-5x4-ASM-PRO \n");
  DemoLogf("                   M = Use Thor-2-PCMCIA-PRO Board\n");
  DemoLogf("                   P = Use Thor-8-PCI-Plus Board with Vidar-5x4-ASM-PRO \n");
  //DemoLogf("                   H = Use Hugin-2-PC104 Board\n");
  DemoLogf(" /L<fileName>  : Log file name\n");
  DemoLogf(" /KI           : Use Internal board oscillator as clock source\n");
  DemoLogf(" /K0           : Use recovered clock from Line Interface 0 as clock source\n");
  DemoLogf(" /K1           : Use recovered clock from Line Interface 1 as clock source\n");
  DemoLogf(" /K3           : Split Clock, both Line Interface are synchronized to its T1E1 span\n");
  DemoLogf(" /K4           : Split Clock, Li0 synchronized to its T1E1 span. Li1 internal\n");
  DemoLogf(" /K5           : Split Clock, Li1 synchronized to its T1E1 span. Li0 internal\n");
  DemoLogf(" /KE           : Use External clock source\n");
  DemoLogf(" /ME           : E1 Li mode\n");
  DemoLogf(" /MT           : T1 Li mode\n");
  DemoLogf(" /D            : DSP Debugger mode. Prompt to start DSP code from Debugger\n");
  DemoLogf(" /F[E|S|9]     : T1 Frame Format, where\n");
  DemoLogf("                   E = ESF (default)\n");
  DemoLogf("                   4 = F4\n");
  DemoLogf("                   S = SF [D3/D4] (F12)\n");
  DemoLogf("                   9 = SLC96 (F72)\n");
  DemoLogf(" /J[A|B]       : T1 Line Code, where\n");
  DemoLogf("                   A = AMI\n");
  DemoLogf("                   B = B8ZS (default)\n");
  DemoLogf(" /G[D|C|G]     : E1 Frame Format, where\n");
  DemoLogf("                   D = Doubleframe\n");
  DemoLogf("                   C = CRC4 Multiframe (default)\n");
  DemoLogf("                   G = CRC4 Multiframe (G.706)\n");
  DemoLogf(" /U[A|H]       : E1 Line Code, where\n");
  DemoLogf("                   A = AMI\n");
  DemoLogf("                   H = HDB3 (default)\n");
  DemoLogf(" /Z<str>       : Command string (overrides keyboard input)\n");
  DemoLogf(" /Q            : Exit without prompting on errors\n");
}


/*******************************************************************************

  @func

  Displays a message on the screen and prints it to the logfile.

  @comm

*******************************************************************************/
static void DemoLogf(char *formatStr, ...)
{
  va_list argList;
  OsaMutexLock(&g_demoLogfCritSec);

  va_start(argList, formatStr);

  OsaConsolePrintF(formatStr, argList);
  if (g_pDemoLogFile) vfprintf(g_pDemoLogFile,   formatStr, argList);

  va_end(argList);

  OsaMutexUnlock(&g_demoLogfCritSec);
}


static void DemoCheckRc(IN OTX_HANDLE hDevice, IN OTX_RESULT result)
/*******************************************************************************
  @func
    Check the result from API call. If the result is a failure code then prompt
    the user if he wants to exit.
*******************************************************************************/
{
  if ( OTX_FAILURE(result) ) {

    if (hDevice == OTX_INVALID_HANDLE_VALUE) {
      DemoLogf("ERROR(%#lX) \n", result);
    } else {
      OTX_RESULT otxRc;
      char msgStr[256];
      OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
      OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
      OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
      OTX_UINT32      eDeviceType = 999;    // Device type for hDevice 
      OTX_UINT32      nDeviceNumber = 999;  // Device number for hDevice
      
      // Get the Device Type for hHandle
      otxRc = OtxDrvGetAttributeType(hDevice, OTX_ATTR_DEVICE_TYPE, &eAttribType, &cbAttribValue, &bReadOnly); 
      if (OTX_SUCCESS(otxRc)) {
        assert(eAttribType == OTX_ATTRIB_UINT32);
        assert(cbAttribValue == sizeof(OTX_UINT32));
        OtxDrvGetAttributeValueUint32(hDevice, OTX_ATTR_DEVICE_TYPE, &eDeviceType); 
      } 
      
      // Get the Device Number for hHandle
      otxRc = OtxDrvGetAttributeType(hDevice, OTX_ATTR_DEVICE_NO, &eAttribType, &cbAttribValue, &bReadOnly); 
      if (OTX_SUCCESS(otxRc)) {
        assert(eAttribType == OTX_ATTRIB_UINT32);
        assert(cbAttribValue == sizeof(OTX_UINT32));
        OtxDrvGetAttributeValueUint32(hDevice, OTX_ATTR_DEVICE_NO, &nDeviceNumber); 
      }
      
      // Convert the result code to a more descriptive string
      otxRc = OtxDrvResultCode2String(hDevice, result, 256, &msgStr[0]);
      if (OTX_FAILURE(otxRc)) {
        strcpy(msgStr, "Cannot translate this result code.");
      }
      DemoLogf("ERROR(%#lX [%d.%d]) Dev(%X,%d,%d): %s \n", OTX_RESULT_CODE(result), OTX_RESULT_SOURCE_ID(result), OTX_RESULT_LINE_NUM(result), hDevice, (int) eDeviceType, (int)nDeviceNumber, msgStr);
    }

    DemoExitQuestion();
  }
}


/*******************************************************************************
  @func
    Prompts whether to exit the program. Usually called after an error 
    has been reported.

*******************************************************************************/
void DemoExitQuestion()
{
  short c;

  if (g_demoEnv.bExitOnError) {
    c = 'y';
  } else {
    DemoLogf("Do you want to exit (Y/N) ");
    c = OsaConsoleGetCh();
  }
  if (c == 'y' || c == 'Y') {
    // Clean up
    OtxDrvReleaseLib();
    OsaConsoleExit();
    exit(1);
  }
}


OTX_INT32 DemoProcessUserCommand(char commandChar)
/*******************************************************************************

  @func

  Processes User Commands

  @comm

*******************************************************************************/
{
  static OTX_INT32 nCurrentLi = 0;
  static OTX_INT32 nCurrentGenerator = 1;
  OTX_RESULT result;

  switch (commandChar) {

  case '?':
  case 'h':
  case 'H':
    DemoLogf("Cmd  Synopsis    (%s)\n", g_demoEnv.eT1E1Mode == OTX_T1E1_MODE_T1 ? "T1" : "E1");
    DemoLogf("===  ========\n");
    DemoLogf(" ?   Display Help text\n");
    DemoLogf(" h   Ditto\n");
    DemoLogf(" q   Quit\n");
    DemoLogf(" 0-1 Select active Li or Dsp, this Li/Dsp is then used for subsequent commands\n");
    DemoLogf(" a   Turn on  Constant value on all timeslots (choice of value)\n");
    DemoLogf(" c   Turn on  Constant value on a selected timeslot\n");
    DemoLogf(" C   Turn off Constant value on a selected timeslot\n");
    DemoLogf(" s   Read sample value from a selected timeslot\n");
    DemoLogf(" S   Read sample value from all timeslots\n");
    DemoLogf(" m   Increment Generator (currently Generator#%d (Li%d, TS%d) Selected\n", nCurrentGenerator, nCurrentGenerator/32, nCurrentGenerator%32);
    DemoLogf(" M   Decrement Generator (currently Generator#%d (Li%d, TS%d) Selected\n", nCurrentGenerator, nCurrentGenerator/32, nCurrentGenerator%32);
    DemoLogf(" l   Query Line Status of both Lis\n");
    break;

  case 'l':
    {
      OtxT1E1Layer1StatusE eLayer1Status;
      OTX_INT32 nLi;

      for (nLi = 0; nLi < 2; nLi++) {
        DemoLogf("Line status of Li #%d is ", nLi);
        result = OtxT1E1GetLayer1Status(g_hPhysDevs.hLi[nLi], &eLayer1Status); DemoCheckRc(g_hPhysDevs.hLi[nLi], result);
        DemoLogf("%s\n", eLayer1Status == OTX_T1E1_LAYER1_OK ? "OK" : "DOWN");
      }
    }
    break;


  case 'a':
    {
      // Turn on Constant Value on all Dialers for this Li
      OTX_UINT16 nConstVal;
      OTX_UINT16 nVal;
      OTX_INT32 i;
      OtxDeviceStateE  eDevState;

      DemoLogf("Li%d: Turn on Constant Value for all Dialers on this Li. \n", nCurrentLi);
      
      DemoLogf("Choose constant value:\n");
      DemoLogf("0 = 0x00\n");
      DemoLogf("1 = 0xFF\n");
      DemoLogf("2 = 0xA5\n");
      DemoLogf("3 = timeslot number (TS0=0, TS1=1, etc)\n");
      switch (OsaConsoleGetCh() - '0') {
      case 0:
        nConstVal = 0;
        break;
      case 1:
        nConstVal = 0xFF;
        break;
      case 2:
        nConstVal = 0xA5;
        break;
      case 3:
        nConstVal = 0xFFFF;
        break;
      default:
        DemoLogf("Invalid entry. Setting constant value to timeslot number\n");
        nConstVal = 0xFFFF;
        break;
      }

      for (i = 0 + 32*nCurrentLi; i < (32 + 32*nCurrentLi); i++) {
        
        if (nConstVal == 0xFFFF) {
          nVal = (OTX_UINT16)i;
        } else {
          nVal = nConstVal;
        }

        result = OtxToneConstValGeneratorInit(g_sqConstValGenerator[i].hHandle, OTX_DATA_FORMAT_UINT8, nVal); DemoCheckRc(g_sqConstValGenerator[i].hHandle, result);
        if (OTX_RESULT_CODE(result) != OTX_S_OK) {
          DemoLogf("Error: Could not initialize ConstValGenerator[%d]\n", i);
          continue;
        }
        
        result = OtxDrvGetState(g_sqConstValGenerator[i].hHandle, &eDevState); DemoCheckRc(g_sqConstValGenerator[i].hHandle, result);
        if (OTX_RESULT_CODE(result) != OTX_S_OK) {
          DemoLogf("Error: Could not get the state for ConstValGenerator[%d]\n", i);
          continue;
        }
        
        //DemoLogf("Enabling ConstValGenerator[%d] device.\n", i);
        if (eDevState != OTX_DEVICE_WORKING) {
          result = OtxDrvEnable(g_sqConstValGenerator[i].hHandle,  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_sqConstValGenerator[i].hHandle, result);
        } else {
          //DemoLogf("Device already enabled. No need to enable it again\n");
        }
        
      }

      DemoLogf("\nNow generating the specified constant values\n");
      
    }
    break;


  case 'c':
    {
      // Turn on Constant Value on the current Generator
      char szBuf[12];
      static OTX_UINT16 nConstVal = 0x55;
      OtxDeviceStateE  eDevState;
                                      
      DemoLogf("Turn on Constant Value on Generator #%d (Li%d, TS%d) \n", nCurrentGenerator, nCurrentGenerator/32, nCurrentGenerator%32);
      DemoLogf("Enter value (0x%.2X)?", nConstVal);
      OsaConsoleGetStr(szBuf);
      if (*szBuf) {
        nConstVal = (OTX_UINT16)strtol(szBuf, NULL, 0);
      }
      
      result = OtxToneConstValGeneratorInit(g_sqConstValGenerator[nCurrentGenerator].hHandle, OTX_DATA_FORMAT_UINT8, nConstVal); DemoCheckRc(g_sqConstValGenerator[nCurrentGenerator].hHandle, result);
      if (OTX_RESULT_CODE(result) != OTX_S_OK) {
        break;
      }
      
      
      result = OtxDrvGetState(g_sqConstValGenerator[nCurrentGenerator].hHandle, &eDevState); DemoCheckRc(g_sqConstValGenerator[nCurrentGenerator].hHandle, result);
      if (OTX_RESULT_CODE(result) != OTX_S_OK) {
        break;
      }
      
      
      DemoLogf("Enabling ConstValGenerator device.\n");
      if (eDevState != OTX_DEVICE_WORKING) {
        result = OtxDrvEnable(g_sqConstValGenerator[nCurrentGenerator].hHandle,  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_sqConstValGenerator[nCurrentGenerator].hHandle, result);
      } else {
        //DemoLogf("Device already enabled. No need to enable it again\n");
      }
      
    }
    break;

  case 'C':
    {
      // Turn off Constant Value on current dialer
      
      DemoLogf("Turn off Constant Value on Generator #%d (Li%d, TS%d)\n", nCurrentGenerator, nCurrentGenerator/32, nCurrentGenerator%32);
      
      result = OtxDrvDisable(g_sqConstValGenerator[nCurrentGenerator].hHandle,  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_sqConstValGenerator[nCurrentGenerator].hHandle, result);
    }
    break;

  case 's':
    {
      // Get sample value from a Generator
      OTX_UINT16 nSampleVal;
                               
      DemoLogf("ValSampler[%d] (Li%d, TS%d): ", nCurrentGenerator, nCurrentGenerator/32, nCurrentGenerator%32);
      
      result = OtxToneValSamplerReadData(g_sqValSampler[nCurrentGenerator].hHandle, &nSampleVal); DemoCheckRc(g_sqValSampler[nCurrentGenerator].hHandle, result);
      
      DemoLogf("0x%.2X \n", nSampleVal);
      
    }
    break;


  case 'S':
    {
      // Get sample value from all Samplers on this Li
      OTX_UINT16 nSampleVal;
      OTX_INT32 i;

      DemoLogf("ValSampler for Li%d: \n", nCurrentLi);
      for (i = 0 + nCurrentLi*32; i < (32 + nCurrentLi*32); i++) {
        if ((i != 0) && ((i%8) == 0)) DemoLogf("\n");
        
        result = OtxToneValSamplerReadData(g_sqValSampler[i].hHandle, &nSampleVal); DemoCheckRc(g_sqValSampler[i].hHandle, result);
        
        DemoLogf("[%.2d]: %.2X ", i%32, nSampleVal);
      }
      
      DemoLogf("\n\n");
      
    }
    break;


  case 'z':
    {
      OtxT1E1TermModeE eTermMode;
      DemoLogf("Set Line Termination mode:\n");
      DemoLogf("0 = OTX_T1E1_TERM_120_OHM\n");
      DemoLogf("1 = OTX_T1E1_TERM_100_OHM\n");
      DemoLogf("2 = OTX_T1E1_TERM_75_OHM\n");
      DemoLogf("3 = OTX_T1E1_TERM_HIGHZ\n");
      switch (OsaConsoleGetCh() - '0') {
      case 0:
        eTermMode = OTX_T1E1_TERM_120_OHM;
        break;
      case 1:
        eTermMode = OTX_T1E1_TERM_100_OHM;
        break;
      case 2:
        eTermMode = OTX_T1E1_TERM_75_OHM;
        break;
      case 3:
        eTermMode = OTX_T1E1_TERM_HIGHZ;
        break;
      default:
        DemoLogf("Invalid entry. Setting Line Termination to OTX_T1E1_TERM_100_OHM\n");
        eTermMode = OTX_T1E1_TERM_100_OHM;
        break;
      }

      result = OtxT1E1SetLineTermination(g_hPhysDevs.hLi[nCurrentLi], eTermMode); DemoCheckRc(g_hPhysDevs.hLi[nCurrentLi], result);
    }
    break;

  case '0':
    // Select Li/Dsp 0 as current
    nCurrentLi = 0;
    DemoLogf("Li/Dsp#0 Selected ....\n");
    break;

  case '1':
    // Select Li/Dsp 1 as current
    nCurrentLi = 1;
    DemoLogf("Li/Dsp#1 Selected ....\n");
    break;


  case 'm':
    {
      // Select current nCurrentGenerator
      nCurrentGenerator++;
      if (nCurrentGenerator > 63) {
        // Wrap 
        nCurrentGenerator = 0;
        nCurrentGenerator++;
      }
      
      DemoLogf("Generator#%d (Li%d, TS%d) Selected \n", nCurrentGenerator, nCurrentGenerator/32, nCurrentGenerator%32);
    }
    break;
    
    
  case 'M':
    {
      // Select current nCurrentGenerator
      nCurrentGenerator--;
      if (nCurrentGenerator < 0) {
        // Wrap 
        nCurrentGenerator = 63;
      }
      
      DemoLogf("Generator#%d (Li%d, TS%d) Selected \n", nCurrentGenerator, nCurrentGenerator/32, nCurrentGenerator%32);
    }
    break;
    
  case '~':
    {
      char strBuf[12];
      OTX_INT32 nSleepTime;

      DemoLogf("Enter milliseconds to sleep :");
      OsaConsoleGetStr(strBuf);

      if (strBuf[0]) {
        nSleepTime = strtol(strBuf, NULL, 0);
        DemoLogf("Sleeping %d milliseconds\n", nSleepTime);
        OsaThreadSleep(nSleepTime);
      }
    }
    break;

  case 'q':
  case 'Q':
    return 1;
    break;

  default:
    DemoLogf("Unknown command. Press '?' for help, 'q' to quit\n");
    break;
  }

  return 0;
}                      


/*******************************************************************************
*******************************************************************************/
void DemoAtExit(void)
{
  fclose(g_pDemoLogFile);
}


int main(int argc, char *argv[])
/*******************************************************************************
*******************************************************************************/
{
  OTX_INT32 exitFlag = 0;
  OTX_RESULT result;  // OTX Standard Return Code
  char commandChar;

	OsaConsoleInit();

  printf("T2MSAMPLES, %s, Odin TeleSystems Inc.\n", __DATE__);
  printf("Vidar-5x4-ASM-PRO/Thor-2-PCMCIA-PRO Example program.\n");

  OsaMutexInit(&g_analyzeCritSec);
  OsaMutexInit(&g_demoLogfCritSec);

  DemoParseCmdLineArgs(argc, argv, &g_demoEnv);

  if (g_demoEnv.pszInputStr) {
    OsaConsoleInput(g_demoEnv.pszInputStr);
  }

  if ((g_pDemoLogFile = fopen(g_demoEnv.logFileName, "wt")) == NULL) {
    printf("Couldn't open logfile %s\n", g_demoEnv.logFileName);
    OsaConsoleExit();
    exit(-1);
  }
  atexit(DemoAtExit);

  {
    char *pszClockSource;

    switch (g_demoEnv.eClkSource) {
    case OTX_CLOCK_SOURCE_INTERNAL:
      pszClockSource = "Internal";
      break;
    case OTX_CLOCK_SOURCE_LOCAL_0:
      pszClockSource = "Line Interface 0";
      break;
    case OTX_CLOCK_SOURCE_LOCAL_1:
      pszClockSource = "Line Interface 1";
      break;
    case OTX_CLOCK_SOURCE_LOCAL_2:
      pszClockSource = "External";
      break;
    case OTX_CLOCK_SOURCE_LOCAL_3:
      pszClockSource = "Split Clock, both Line Interface are synchronized to its T1E1 span";
      break;
    case OTX_CLOCK_SOURCE_LOCAL_4:
      pszClockSource = "Split Clock, Li0 synchronized to its T1E1 span. Li1 internal";
      break;
    case OTX_CLOCK_SOURCE_LOCAL_5:
      pszClockSource = "Split Clock, Li1 synchronized to its T1E1 span. Li0 internal";
      break;
    default:
      pszClockSource = "Unknown";
      break;
    }

    DemoLogf("Log file is %s\n", g_demoEnv.logFileName);
    DemoLogf("Clock Source: %s\n", pszClockSource);
  }

  // Connect to the Driver
  DemoLogf("Connecting lib...\n");
  result = OtxDrvConnectLib();
  switch (OTX_RESULT_CODE(result)) {
  case OTX_RES_HANDLE:
    DemoLogf("Error: Could not connect to the OTX library.\nPossible cause: OTX Driver not loaded\n");
    OsaConsoleExit();
    OtxDrvReleaseLib();
    exit(1);
    break;
  case OTX_RES_NOT_SUPPORTED:
    DemoLogf("Error: Mismatch between OTX Driver revision and OTX Library revision.\nRevisions must match.\n");
    OsaConsoleExit();
    OtxDrvReleaseLib();
    exit(1);
    break;
  default:
    DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    break;
  }

  // Query driver revision
  {
    char szVersionInfo[255];
    result = OtxDrvGetAttributeValueSz(OTX_DRIVER_DEVICE_HANDLE, OTX_ATTR_DRV_REV_STR, szVersionInfo, 255); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);

    DemoLogf("Driver version %s\n", szVersionInfo);
  }

  // Create Event Queues for Driver --> Application Communications
  DemoLogf("DemoCreateEventQueue\n");
  DemoCreateEventQueue();

  // Connect (get handles) to the physical devices on the Thor-2-PCMCIA board
  DemoLogf("DemoCreatePhysicalDevices\n");
  DemoCreatePhysicalDevices();

  // Read Serial Number from the board
  {
    OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
    OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
    OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
    char *pszSerialNumber;

    result = OtxDrvGetAttributeType(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_SERIAL_NUMBER, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(g_hPhysDevs.hThorController0, result);
    assert(eAttribType == OTX_ATTRIB_SZ);
    assert(bReadOnly == OTX_TRUE);
    pszSerialNumber = (char *)malloc(cbAttribValue * sizeof(char));
    result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_SERIAL_NUMBER, pszSerialNumber, cbAttribValue); DemoCheckRc(g_hPhysDevs.hThorController0, result);
    DemoLogf("Board Serial Number: %s\n", pszSerialNumber);
    free(pszSerialNumber);
  }

  // Read Build Number from the board
  {
    OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
    OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
    OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
    char *pszBrdBuildNumber;

    result = OtxDrvGetAttributeType(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_BUILD, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(g_hPhysDevs.hThorController0, result);
    assert(eAttribType == OTX_ATTRIB_SZ);
    assert(bReadOnly == OTX_TRUE);
    pszBrdBuildNumber = (char *)malloc(cbAttribValue * sizeof(char));
    result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_BUILD, pszBrdBuildNumber, cbAttribValue); DemoCheckRc(g_hPhysDevs.hThorController0, result);
    DemoLogf("Board Build Number: %s\n", pszBrdBuildNumber);
    free(pszBrdBuildNumber);
  }

  // Read User String from the board
  {
    OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
    OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
    OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
    char *pszBrdUserString;

    result = OtxDrvGetAttributeType(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_USER_STRING, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(g_hPhysDevs.hThorController0, result);
    assert(eAttribType == OTX_ATTRIB_SZ);
    assert(bReadOnly == OTX_TRUE);
    pszBrdUserString = (char *)malloc(cbAttribValue * sizeof(char));
    result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_USER_STRING, pszBrdUserString, cbAttribValue); DemoCheckRc(g_hPhysDevs.hThorController0, result);
    DemoLogf("User String: %s\n", pszBrdUserString);
    free(pszBrdUserString);
  }

  // Read Bus Address from the board
  {
    OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
    OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
    OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
    OTX_UINT32      nBusAddress;
    
    result = OtxDrvGetAttributeType(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_BUS_ADDRESS, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(g_hPhysDevs.hThorController0, result);
    assert(eAttribType == OTX_ATTRIB_UINT32);
    assert(bReadOnly == OTX_TRUE);
    assert(cbAttribValue == sizeof(OTX_UINT32));
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_BUS_ADDRESS, &nBusAddress); DemoCheckRc(g_hPhysDevs.hThorController0, result);
    DemoLogf("Bus Address: 0x%X\n", nBusAddress);
  }


  // Read Pnp Id from the board
  {
    OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
    OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
    OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
    OTX_UINT32      cbwcszPnpId;
    wchar_t *pwcszPnpId;
    
    result = OtxDrvGetAttributeType(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_PNP_ID, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(g_hPhysDevs.hThorController0, result);
    assert(eAttribType == OTX_ATTRIB_OCTETS); // We read the PnP Id as a string of octets from the OTX driver, but the value is actually a wide-character string.
    assert(bReadOnly == OTX_TRUE);
    assert((cbAttribValue % 2) == 0); 
    if (cbAttribValue) {
      pwcszPnpId = (wchar_t *)malloc(cbAttribValue);
      result = OtxDrvGetAttributeValue(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_PNP_ID, (OTX_PVOID)pwcszPnpId, cbAttribValue, &cbwcszPnpId); DemoCheckRc(g_hPhysDevs.hThorController0, result);
      DemoLogf("PnP Id: %S\n", pwcszPnpId);
      free(pwcszPnpId);
    } else {
      DemoLogf("PnP Id: None\n");
    }
  }


  switch (g_demoEnv.nBoardType) {
  case OTX_DEVICE_HUGIN_2_PC104:
  case OTX_DEVICE_THOR_2_PCI_PLUS:
  case OTX_DEVICE_THOR_4_PCI_PLUS:
  case OTX_DEVICE_THOR_8_PCI_PLUS:
    // Query PCI Device and PCI Function Number for this PCI board
    {
      OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
      OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
      OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
      OTX_UINT32      nPciDeviceNumber;
      OTX_UINT32      nPciFunctionNumber;
      
      result = OtxDrvGetAttributeType(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_PCI_DEVICE_NUMBER, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(g_hPhysDevs.hThorController0, result);
      assert(eAttribType == OTX_ATTRIB_UINT32);
      assert(bReadOnly == OTX_TRUE);
      assert(cbAttribValue == sizeof(OTX_UINT32));
      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_PCI_DEVICE_NUMBER, &nPciDeviceNumber); DemoCheckRc(g_hPhysDevs.hThorController0, result);

      DemoLogf("PCI Device Number : %.4X\n", nPciDeviceNumber);

      result = OtxDrvGetAttributeType(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_PCI_FUNCTION_NUMBER, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(g_hPhysDevs.hThorController0, result);
      assert(eAttribType == OTX_ATTRIB_UINT32);
      assert(bReadOnly == OTX_TRUE);
      assert(cbAttribValue == sizeof(OTX_UINT32));
      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_PCI_FUNCTION_NUMBER, &nPciFunctionNumber); DemoCheckRc(g_hPhysDevs.hThorController0, result);

      DemoLogf("PCI Function Number: %.4X\n", nPciFunctionNumber);

    }
    break;
  case OTX_DEVICE_THOR_PCMCIA:
    // Query PCMCIA Socket Number for this PCMCIA board
    {
      OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
      OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
      OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
      OTX_UINT32      nPcmciaSocketNumber;
      
      result = OtxDrvGetAttributeType(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_PCMCIA_SOCKET_NUMBER, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(g_hPhysDevs.hThorController0, result);
      assert(eAttribType == OTX_ATTRIB_UINT32);
      assert(bReadOnly == OTX_TRUE);
      assert(cbAttribValue == sizeof(OTX_UINT32));
      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_PCMCIA_SOCKET_NUMBER, &nPcmciaSocketNumber); DemoCheckRc(g_hPhysDevs.hThorController0, result);
      DemoLogf("PCMCIA Socket Number: %d\n", nPcmciaSocketNumber);
    }

    // Query Device ID for this PCMCIA board to determine which type of PCMCIA card it is
    {
      OTX_UINT32      nDeviceId;
      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_DEVICE_ID, &nDeviceId); DemoCheckRc(g_hPhysDevs.hThorController0, result);
      switch (nDeviceId) {
      case 1060: DemoLogf("Thor-2-PCMCIA-PRO card detected\n"); break;
      case 1058: DemoLogf("Thor-2-PCMCIA-EX card detected\n"); break;
      case 1057: DemoLogf("Thor-2-PCMCIA-CST card detected\n"); break;
      case 1043: DemoLogf("Thor-2-PCMCIA-PLUS card detected\n"); break;
      case 1012: DemoLogf("Thor-2-PCMCIA-STD card detected\n"); break;
      default: DemoLogf("Error: Unknown Thor-2-PCMCIA card (deviceId=%d)\n", nDeviceId); break;
      }
      
      if (nDeviceId != 1060) { 
        DemoLogf("Only the Thor-2-PCMCIA-PRO card variant is supported by this demo program\n");
        exit(1);
      }
      
    }
    break;

  default:
    DemoLogf("Board Type (%d) not supported by this demo program\n", g_demoEnv.nBoardType);
    exit(1);
    break;
  }


  // Configure E1/T1 devices
  DemoLogf("Configuring E1/T1 devices...\n");
  DemoConfigureE1T1Devices();
  DemoLogf("Configured E1/T1 devices\n");

  // Enable physical devices. Their device state should now be OTX_DEVICE_WORKING
  DemoLogf("DemoEnablePhysicalDevices\n");
  DemoEnablePhysicalDevices();

  // Set clock source for the board
  if (g_hPhysDevs.hVidarAsmController0) {
    // Thor-2-PCI
    result = OtxBrdSetClocks(g_hPhysDevs.hThorController0, OTX_CLOCK_MODE_LOCAL, g_demoEnv.eClkSource); DemoCheckRc(g_hPhysDevs.hThorController0, result);
  } else {
    // Thor-2-PCMCIA
    result = OtxBrdSetClocks(g_hPhysDevs.hThorController0, OTX_CLOCK_MODE_NA, g_demoEnv.eClkSource); DemoCheckRc(g_hPhysDevs.hThorController0, result);
  }


  // Create and get handles to the Logical devices needed by this application
  DemoLogf("DemoCreateLogicalDevices\n");
  DemoCreateLogicalDevices();

  // Create and get handles to the Pipe devices needed by this application
  DemoLogf("DemoConnectDevices\n");
  DemoConnectDevices();

  // Initialize the logical devices. A logical device must be create and connected before it can be initialized.
  DemoLogf("DemoInitializeLogicalDevices\n");
  DemoInitializeLogicalDevices();

  // Enable logical devices. Their device state should now be OTX_DEVICE_WORKING
  DemoLogf("DemoEnableLogicalDevices\n");
  DemoEnableLogicalDevices();



  // Main Loop. Loop here and wait for Events from the driver or commands from the user ....
	OsaThreadStart(DemoIdleFunction);

  OsaThreadSleep(300); // Only so the following message will show on the screen

  DemoLogf("\nPress 'h' for help\n");

  while (!exitFlag) {
    commandChar = OsaConsoleGetCh();
    exitFlag = DemoProcessUserCommand(commandChar);
  }


  OsaConsoleExit();
  OtxDrvReleaseLib();

  return 0;
}

#if defined(_WIN32_WCE)
int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPWSTR lpCmdLine, int nShowCmd) 
{
  // default command line parameters
  WCHAR wsExeName[OTX_MAX_PATH * sizeof(WCHAR)];
  static char exename[OTX_MAX_PATH];
  static char *cmdParams[] = { exename, "/TM", NULL }; 
  char *pc;

  GetModuleFileName(NULL, wsExeName, sizeof(wsExeName));
  WideCharToMultiByte(CP_ACP, 0, &wsExeName[0], sizeof(wsExeName), exename, sizeof(exename), NULL, NULL); 
  OsaStrCpy(g_demoEnv.cwd, cmdParams[0]);
  pc = OsaStrRChr(g_demoEnv.cwd, '\\');
  if (pc != NULL) {
    *++pc = '\0';
  }

  return OsaWinCeConsoleMain(hInstance, hPrevInstance, lpCmdLine, nShowCmd,
              cmdParams, sizeof(cmdParams) / sizeof(cmdParams[0]), main);
}
#endif  // #if defined(_WIN32_WCE)
